#include <stdio.h>
#include "cuda.h"
#define max(x,y)  ((x) > (y)? (x) : (y))
#define min(x,y)  ((x) < (y)? (x) : (y))
#define ceil(a,b) ((a) % (b) == 0 ? (a) / (b) : ((a) / (b)) + 1)

void check_error (const char* message) {
	cudaError_t error = cudaGetLastError ();
	if (error != cudaSuccess) {
		printf ("CUDA error : %s, %s\n", message, cudaGetErrorString (error));
		exit(-1);
	}
}

__global__ void j2d64pt (double * __restrict__ l_in, double * __restrict__ l_out, int N) {
	//Determing the block's indices
	int i0 = (int)(blockIdx.x)*(int)(blockDim.x) + 4;
	int i = max(i0,4) + (int)(threadIdx.x);
	int j0 = 4*(int)(blockIdx.y)*(int)(blockDim.y) + 4;
	int j = max(j0,4) + 4*(int)(threadIdx.y);

	double (*in)[8200] = (double (*)[8200]) l_in;
	double (*out)[8200] = (double (*)[8200]) l_out;

	if (i>=4 & j>=4 & i<=N-5 & j<=N-5) {
double outjc0ic0 = 1.27449 * in[j-4][i-4];
outjc0ic0 -= 1.27449 * in[j-4][i+4];
outjc0ic0 += 0.000136017 * -in[j-4][i-3];
outjc0ic0 += 0.000136017 * in[j-4][i+3];
outjc0ic0 += 0.000714 * in[j-4][i-2];
outjc0ic0 -= 0.000714 * in[j-4][i+2];
outjc0ic0 += 0.002856 * -in[j-4][i-1];
outjc0ic0 += 0.002856 * in[j-4][i+1];
outjc0ic0 -= 1.27449 * in[j+4][i-4];
double outjp1ic0 = 0.000136017 * in[j+4][i-4];
double outjp2ic0 = -(0.000714 * in[j+4][i-4]);
double outjp3ic0 = 0.002856 * in[j+4][i-4];
outjc0ic0 += 1.27449 * in[j+4][i+4];
outjp1ic0 -= 0.000136017 * in[j+4][i+4];
outjp2ic0 += 0.000714 * in[j+4][i+4];
outjp3ic0 -= 0.002856 * in[j+4][i+4];
outjc0ic0 += 0.000136017 * in[j-3][i+4];
outjp1ic0 -= 1.27449 * in[j-3][i+4];
outjc0ic0 -= 0.000136017 * in[j-3][i-4];
outjp1ic0 += 1.27449 * in[j-3][i-4];
outjc0ic0 += 0.000136017 * in[j+3][i-4];
outjp1ic0 -= 0.000714 * in[j+3][i-4];
outjp2ic0 += 0.002856 * in[j+3][i-4];
outjc0ic0 -= 0.000136017 * in[j+3][i+4];
outjp1ic0 += 0.000714 * in[j+3][i+4];
outjp2ic0 -= 0.002856 * in[j+3][i+4];
outjc0ic0 += 0.000136017 * in[j+4][i-3];
outjp1ic0 -= 0.00145161 * in[j+4][i-3];
outjp2ic0 += 0.00762 * in[j+4][i-3];
outjp3ic0 -= 0.03048 * in[j+4][i-3];
outjc0ic0 -= 0.000136017 * in[j+4][i+3];
outjp1ic0 += 0.00145161 * in[j+4][i+3];
outjp2ic0 -= 0.00762 * in[j+4][i+3];
outjp3ic0 += 0.03048 * in[j+4][i+3];
outjc0ic0 += 0.000714 * in[j-2][i-4];
outjp1ic0 -= 0.000136017 * in[j-2][i-4];
outjp2ic0 += 1.27449 * in[j-2][i-4];
outjc0ic0 -= 0.000714 * in[j-2][i+4];
outjp1ic0 += 0.000136017 * in[j-2][i+4];
outjp2ic0 -= 1.27449 * in[j-2][i+4];
outjc0ic0 -= 0.000714 * in[j+2][i-4];
outjp1ic0 += 0.002856 * in[j+2][i-4];
outjp3ic0 -= 0.002856 * in[j+2][i-4];
outjc0ic0 += 0.000714 * in[j+2][i+4];
outjp1ic0 -= 0.002856 * in[j+2][i+4];
outjp3ic0 += 0.002856 * in[j+2][i+4];
outjc0ic0 -= 0.000714 * in[j+4][i-2];
outjp1ic0 += 0.00762 * in[j+4][i-2];
outjp2ic0 -= 0.04 * in[j+4][i-2];
outjp3ic0 += 0.16 * in[j+4][i-2];
outjc0ic0 += 0.000714 * in[j+4][i+2];
outjp1ic0 -= 0.00762 * in[j+4][i+2];
outjp2ic0 += 0.04 * in[j+4][i+2];
outjp3ic0 -= 0.16 * in[j+4][i+2];
outjc0ic0 -= 0.002856 * in[j-1][i-4];
outjp1ic0 += 0.000714 * in[j-1][i-4];
outjp2ic0 -= 0.000136017 * in[j-1][i-4];
outjp3ic0 += 1.27449 * in[j-1][i-4];
outjc0ic0 += 0.002856 * in[j-1][i+4];
outjp1ic0 -= 0.000714 * in[j-1][i+4];
outjp2ic0 += 0.000136017 * in[j-1][i+4];
outjp3ic0 -= 1.27449 * in[j-1][i+4];
outjc0ic0 += 0.002856 * in[j+1][i-4];
outjp2ic0 -= 0.002856 * in[j+1][i-4];
outjp3ic0 += 0.000714 * in[j+1][i-4];
outjc0ic0 -= 0.002856 * in[j+1][i+4];
outjp2ic0 += 0.002856 * in[j+1][i+4];
outjp3ic0 -= 0.000714 * in[j+1][i+4];
outjc0ic0 += 0.002856 * in[j+4][i-1];
outjp1ic0 -= 0.03048 * in[j+4][i-1];
outjp2ic0 += 0.16 * in[j+4][i-1];
outjp3ic0 -= 0.64 * in[j+4][i-1];
outjc0ic0 -= 0.002856 * in[j+4][i+1];
outjp1ic0 += 0.03048 * in[j+4][i+1];
outjp2ic0 -= 0.16 * in[j+4][i+1];
outjp3ic0 += 0.64 * in[j+4][i+1];
outjc0ic0 += 0.00145161 * in[j-3][i-3];
outjp1ic0 += 0.000136017 * -in[j-3][i-3];
outjc0ic0 -= 0.00145161 * in[j-3][i+3];
outjp1ic0 += 0.000136017 * in[j-3][i+3];
outjc0ic0 -= 0.00145161 * in[j+3][i-3];
outjp1ic0 += 0.00762 * in[j+3][i-3];
outjp2ic0 -= 0.03048 * in[j+3][i-3];
outjc0ic0 += 0.00145161 * in[j+3][i+3];
outjp1ic0 -= 0.00762 * in[j+3][i+3];
outjp2ic0 += 0.03048 * in[j+3][i+3];
outjc0ic0 += 0.00762 * -in[j-3][i-2];
outjp1ic0 += 0.000714 * in[j-3][i-2];
outjc0ic0 += 0.00762 * in[j-3][i+2];
outjp1ic0 -= 0.000714 * in[j-3][i+2];
outjc0ic0 -= 0.00762 * in[j-2][i-3];
outjp1ic0 += 0.00145161 * in[j-2][i-3];
outjp2ic0 += 0.000136017 * -in[j-2][i-3];
outjc0ic0 += 0.00762 * in[j-2][i+3];
outjp1ic0 -= 0.00145161 * in[j-2][i+3];
outjp2ic0 += 0.000136017 * in[j-2][i+3];
outjc0ic0 += 0.00762 * in[j+2][i-3];
outjp1ic0 -= 0.03048 * in[j+2][i-3];
outjp3ic0 += 0.03048 * in[j+2][i-3];
outjc0ic0 -= 0.00762 * in[j+2][i+3];
outjp1ic0 += 0.03048 * in[j+2][i+3];
outjp3ic0 -= 0.03048 * in[j+2][i+3];
outjc0ic0 += 0.00762 * in[j+3][i-2];
outjp1ic0 -= 0.04 * in[j+3][i-2];
outjp2ic0 += 0.16 * in[j+3][i-2];
outjc0ic0 -= 0.00762 * in[j+3][i+2];
outjp1ic0 += 0.04 * in[j+3][i+2];
outjp2ic0 -= 0.16 * in[j+3][i+2];
outjc0ic0 += 0.03048 * in[j-3][i-1];
outjp1ic0 += 0.002856 * -in[j-3][i-1];
outjc0ic0 -= 0.03048 * in[j-3][i+1];
outjp1ic0 += 0.002856 * in[j-3][i+1];
outjc0ic0 += 0.03048 * in[j-1][i-3];
outjp1ic0 -= 0.00762 * in[j-1][i-3];
outjp2ic0 += 0.00145161 * in[j-1][i-3];
outjp3ic0 += 0.000136017 * -in[j-1][i-3];
outjc0ic0 -= 0.03048 * in[j-1][i+3];
outjp1ic0 += 0.00762 * in[j-1][i+3];
outjp2ic0 -= 0.00145161 * in[j-1][i+3];
outjp3ic0 += 0.000136017 * in[j-1][i+3];
outjc0ic0 -= 0.03048 * in[j+1][i-3];
outjp2ic0 += 0.03048 * in[j+1][i-3];
outjp3ic0 -= 0.00762 * in[j+1][i-3];
outjc0ic0 += 0.03048 * in[j+1][i+3];
outjp2ic0 -= 0.03048 * in[j+1][i+3];
outjp3ic0 += 0.00762 * in[j+1][i+3];
outjc0ic0 -= 0.03048 * in[j+3][i-1];
outjp1ic0 += 0.16 * in[j+3][i-1];
outjp2ic0 -= 0.64 * in[j+3][i-1];
outjc0ic0 += 0.03048 * in[j+3][i+1];
outjp1ic0 -= 0.16 * in[j+3][i+1];
outjp2ic0 += 0.64 * in[j+3][i+1];
outjc0ic0 += 0.04 * in[j-2][i-2];
outjp1ic0 += 0.00762 * -in[j-2][i-2];
outjp2ic0 += 0.000714 * in[j-2][i-2];
outjc0ic0 -= 0.04 * in[j-2][i+2];
outjp1ic0 += 0.00762 * in[j-2][i+2];
outjp2ic0 -= 0.000714 * in[j-2][i+2];
outjc0ic0 -= 0.04 * in[j+2][i-2];
outjp1ic0 += 0.16 * in[j+2][i-2];
outjp3ic0 -= 0.16 * in[j+2][i-2];
outjc0ic0 += 0.04 * in[j+2][i+2];
outjp1ic0 -= 0.16 * in[j+2][i+2];
outjp3ic0 += 0.16 * in[j+2][i+2];
outjc0ic0 += 0.16 * -in[j-2][i-1];
outjp1ic0 += 0.03048 * in[j-2][i-1];
outjp2ic0 += 0.002856 * -in[j-2][i-1];
outjc0ic0 += 0.16 * in[j-2][i+1];
outjp1ic0 -= 0.03048 * in[j-2][i+1];
outjp2ic0 += 0.002856 * in[j-2][i+1];
outjc0ic0 -= 0.16 * in[j-1][i-2];
outjp1ic0 += 0.04 * in[j-1][i-2];
outjp2ic0 += 0.00762 * -in[j-1][i-2];
outjp3ic0 += 0.000714 * in[j-1][i-2];
outjc0ic0 += 0.16 * in[j-1][i+2];
outjp1ic0 -= 0.04 * in[j-1][i+2];
outjp2ic0 += 0.00762 * in[j-1][i+2];
outjp3ic0 -= 0.000714 * in[j-1][i+2];
outjc0ic0 += 0.16 * in[j+1][i-2];
outjp2ic0 -= 0.16 * in[j+1][i-2];
outjp3ic0 += 0.04 * in[j+1][i-2];
outjc0ic0 -= 0.16 * in[j+1][i+2];
outjp2ic0 += 0.16 * in[j+1][i+2];
outjp3ic0 -= 0.04 * in[j+1][i+2];
outjc0ic0 += 0.16 * in[j+2][i-1];
outjp1ic0 -= 0.64 * in[j+2][i-1];
outjp3ic0 += 0.64 * in[j+2][i-1];
outjc0ic0 -= 0.16 * in[j+2][i+1];
outjp1ic0 += 0.64 * in[j+2][i+1];
outjp3ic0 -= 0.64 * in[j+2][i+1];
outjc0ic0 += 0.64 * in[j-1][i-1];
outjp1ic0 += 0.16 * -in[j-1][i-1];
outjp2ic0 += 0.03048 * in[j-1][i-1];
outjp3ic0 += 0.002856 * -in[j-1][i-1];
outjc0ic0 -= 0.64 * in[j-1][i+1];
outjp1ic0 += 0.16 * in[j-1][i+1];
outjp2ic0 -= 0.03048 * in[j-1][i+1];
outjp3ic0 += 0.002856 * in[j-1][i+1];
outjc0ic0 -= 0.64 * in[j+1][i-1];
outjp2ic0 += 0.64 * in[j+1][i-1];
outjp3ic0 += 0.16 * -in[j+1][i-1];
outjc0ic0 += 0.64 * in[j+1][i+1];
outjp2ic0 -= 0.64 * in[j+1][i+1];
outjp3ic0 += 0.16 * in[j+1][i+1];

outjp1ic0 -= 1.27449 * in[j+5][i-4];
outjp2ic0 += 0.000136017 * in[j+5][i-4];
outjp3ic0 -= 0.000714 * in[j+5][i-4];
outjp1ic0 += 1.27449 * in[j+5][i+4];
outjp2ic0 -= 0.000136017 * in[j+5][i+4];
outjp3ic0 += 0.000714 * in[j+5][i+4];
outjp1ic0 += 0.000136017 * in[j+5][i-3];
outjp2ic0 -= 0.00145161 * in[j+5][i-3];
outjp3ic0 += 0.00762 * in[j+5][i-3];
outjp1ic0 -= 0.000136017 * in[j+5][i+3];
outjp2ic0 += 0.00145161 * in[j+5][i+3];
outjp3ic0 -= 0.00762 * in[j+5][i+3];
outjp1ic0 -= 0.000714 * in[j+5][i-2];
outjp2ic0 += 0.00762 * in[j+5][i-2];
outjp3ic0 -= 0.04 * in[j+5][i-2];
outjp1ic0 += 0.000714 * in[j+5][i+2];
outjp2ic0 -= 0.00762 * in[j+5][i+2];
outjp3ic0 += 0.04 * in[j+5][i+2];
outjp1ic0 -= 0.002856 * in[j][i-4];
outjp2ic0 += 0.000714 * in[j][i-4];
outjp3ic0 -= 0.000136017 * in[j][i-4];
outjp1ic0 += 0.002856 * in[j][i+4];
outjp2ic0 -= 0.000714 * in[j][i+4];
outjp3ic0 += 0.000136017 * in[j][i+4];
outjp1ic0 += 0.002856 * in[j+5][i-1];
outjp2ic0 -= 0.03048 * in[j+5][i-1];
outjp3ic0 += 0.16 * in[j+5][i-1];
outjp1ic0 -= 0.002856 * in[j+5][i+1];
outjp2ic0 += 0.03048 * in[j+5][i+1];
outjp3ic0 -= 0.16 * in[j+5][i+1];
outjp1ic0 += 0.03048 * in[j][i-3];
outjp2ic0 -= 0.00762 * in[j][i-3];
outjp3ic0 += 0.00145161 * in[j][i-3];
outjp1ic0 -= 0.03048 * in[j][i+3];
outjp2ic0 += 0.00762 * in[j][i+3];
outjp3ic0 -= 0.00145161 * in[j][i+3];
outjp1ic0 -= 0.16 * in[j][i-2];
outjp2ic0 += 0.04 * in[j][i-2];
outjp3ic0 += 0.00762 * -in[j][i-2];
outjp1ic0 += 0.16 * in[j][i+2];
outjp2ic0 -= 0.04 * in[j][i+2];
outjp3ic0 += 0.00762 * in[j][i+2];
outjp1ic0 += 0.64 * in[j][i-1];
outjp2ic0 += 0.16 * -in[j][i-1];
outjp3ic0 += 0.03048 * in[j][i-1];
outjp1ic0 -= 0.64 * in[j][i+1];
outjp2ic0 += 0.16 * in[j][i+1];
outjp3ic0 -= 0.03048 * in[j][i+1];

outjp2ic0 -= 1.27449 * in[j+6][i-4];
outjp3ic0 += 0.000136017 * in[j+6][i-4];
outjp2ic0 += 1.27449 * in[j+6][i+4];
outjp3ic0 -= 0.000136017 * in[j+6][i+4];
outjp2ic0 += 0.000136017 * in[j+6][i-3];
outjp3ic0 -= 0.00145161 * in[j+6][i-3];
outjp2ic0 -= 0.000136017 * in[j+6][i+3];
outjp3ic0 += 0.00145161 * in[j+6][i+3];
outjp2ic0 -= 0.000714 * in[j+6][i-2];
outjp3ic0 += 0.00762 * in[j+6][i-2];
outjp2ic0 += 0.000714 * in[j+6][i+2];
outjp3ic0 -= 0.00762 * in[j+6][i+2];
outjp2ic0 += 0.002856 * in[j+6][i-1];
outjp3ic0 -= 0.03048 * in[j+6][i-1];
outjp2ic0 -= 0.002856 * in[j+6][i+1];
outjp3ic0 += 0.03048 * in[j+6][i+1];

outjp3ic0 -= 1.27449 * in[j+7][i-4];
outjp3ic0 += 1.27449 * in[j+7][i+4];
outjp3ic0 += 0.000136017 * in[j+7][i-3];
outjp3ic0 -= 0.000136017 * in[j+7][i+3];
outjp3ic0 -= 0.000714 * in[j+7][i-2];
outjp3ic0 += 0.000714 * in[j+7][i+2];
outjp3ic0 += 0.002856 * in[j+7][i-1];
outjp3ic0 -= 0.002856 * in[j+7][i+1];

out[j][i] = outjc0ic0;
out[j+1][i] = outjp1ic0;
out[j+2][i] = outjp2ic0;
out[j+3][i] = outjp3ic0;
	} 
}

extern "C" void host_code (double *h_in, double *h_out, int N) {
	double *in;
	cudaMalloc (&in, sizeof(double)*N*N);
	check_error ("Failed to allocate device memory for in\n");
	cudaMemcpy (in, h_in, sizeof(double)*N*N, cudaMemcpyHostToDevice);
	double *out;
	cudaMalloc (&out, sizeof(double)*N*N);
	check_error ("Failed to allocate device memory for out\n");

	dim3 blockconfig (16, 8);
	dim3 gridconfig (ceil(N, blockconfig.x), ceil(N, 4*blockconfig.y));

	j2d64pt<<<gridconfig, blockconfig>>> (in, out, N);

	cudaMemcpy (h_out, out, sizeof(double)*N*N, cudaMemcpyDeviceToHost);
	cudaFree (in); 
	cudaFree (out);
}
